import numpy as np


class for_w:
    """
           wの極値探索のためのクラス型
    """
    def __init__(self, dist2, n, p, r, x):
        self.dist2 = dist2
        self.n = n
        self.p = p
        self.r = r
        self.x = x    #   xは固定

    def f(self, z):
        """
                １次元配列（ベクトル）zを2次元配列wに並べ換える
        """
        w = np.resize(z, (self.p, self.r))
        w = w**2
        d2_hat = []
        for s in range(self.p):
            d2_temp = np.zeros((self.n,self.n))
            v = self.x * (w[s]**0.5)
            for i in range(self.n):
                for j in range(self.n):
                    d2_temp[i][j] = np.sum((v[i] - v[j])**2)
            d2_hat.append(d2_temp)

        d2_hat = np.array(d2_hat)
        err = np.sum((self.dist2 - d2_hat)**2)

        return err
        
            

class for_x:
    """
            xの極値探索のためのクラス型
    """
    def __init__(self, dist2, n, p, r, w):
        self.dist2 = dist2
        self.n = n
        self.p = p
        self.r = r
        self.w = w    #  wは固定

    def f(self, z):
        """
                １次元配列（ベクトル）zを２次元配列xに並べ換える
        """
        x = np.resize(z, (self.n, self.r))
        d2_hat = []
        for s in range(self.p):
            d2_temp = np.zeros((self.n,self.n))
            v = x * (self.w[s]**0.5)
            for i in range(self.n):
                for j in range(self.n):
                    d2_temp[i][j] = np.sum((v[i] - v[j])**2)
            d2_hat.append(d2_temp)

        d2_hat = np.array(d2_hat)
        err = np.sum((self.dist2 - d2_hat)**2)

        return err

